"""Main module."""
from __future__ import annotations

from functools import reduce

from numpy import exp, zeros
from numpy.random import normal

DEFAULT_NU = 3
DEFAULT_LAMBDA = 1
DEFAULT_SIGMA = 0.5
DEFAULT_ZETA = 0
DEFAULT_ITERANTION_COUNT = 1000


def bary_batch(oracle, xs, nu=DEFAULT_NU):
    """
    Batch barycenter algorithm for direct optimization

    In:
      - oracle     [function]   : Oracle map e.g. lambda x: numpy.norm(x)
      - xs         [list[list]] : list with coordinates
      - nu         [double]     : positive value (Caution due overflow)
      - lambda     [double]     : Forgetting factor between 0 and 1
    Out:
       - xhat      [np.array]   : barycenter position
    """

    if not isinstance(oracle(xs[0]), float):
        match_str = "Oracle function must evaluate as a scalar value."
        raise ValueError(match_str)

    n = len(xs[0])
    size_x = (n, 1)

    def bexp_fun(x):
        return exp(-nu * oracle(x))

    def prod_func(elems):
        return elems[0] * elems[1]

    def sum_func(acc, a):
        return acc + a

    coord_value_iter = zip(map(bexp_fun, xs), xs)
    num = reduce(sum_func, map(prod_func, coord_value_iter), zeros(size_x).T)

    den = reduce(sum_func, map(bexp_fun, xs), 0)

    return num / den


def bary_recur_formula(m_1, xhat_1, x, oracle, nu, lambda_):
    e_i = exp(-nu * oracle(x))
    m = lambda_ * m_1 + e_i

    return m, (1 / m) * (lambda_ * m_1 * xhat_1 + x * e_i)


def bary_recursive(
    oracle,
    x0,
    nu=DEFAULT_NU,
    sigma=DEFAULT_SIGMA,
    zeta=DEFAULT_ZETA,
    lambda_=DEFAULT_LAMBDA,
    iterations=DEFAULT_ITERANTION_COUNT,
):
    """
    Recursive barycenter algorithm for direct optimization

    In:
      - oracle     [function]  : Oracle map e.g. lambda x: numpy.norm(x, 2)
      - x0         [np.array]  : Initial query values
      - nu         [double]    : positive value (Caution due overflow)
      - sigma      [double]    : Std deviation of normal distribution
      - zeta       [double]    : scaler for mean of normal distribution
      - lambda     [double]    : Forgetting factor between 0 and 1
      - iterations [integer]   : Maximum number of iterations

    Out:
       - xhat      [np.array]  : barycenter position
    """

    if not isinstance(oracle(x0), float):
        match_str = "Oracle function must evaluate as a scalar value."
        raise ValueError(match_str)

    # Initialization
    xhat_1 = x0
    m_1 = 0
    card_x = (len(x0), 1)

    deltax_1 = zeros(card_x).T
    solution_is_found = False

    # Optimization loop
    i = 1
    while not solution_is_found:
        z = normal(-zeta * deltax_1, sigma)

        x = xhat_1 + z
        m, xhat = bary_recur_formula(m_1, xhat_1, x, oracle, nu, lambda_)

        # Update previous variables
        m_1, xhat_1, deltax_1 = m, xhat, xhat - xhat_1

        solution_is_found = i >= iterations
        i = i + 1

    return xhat



""" Package pybary  """

# -*- coding: utf-8 -*-

from __future__ import annotations

from .pybary import bary_batch, bary_recursive

__author__ = "Bruno Peixoto"
__email__ = "brunolnetto@gmail.com"

__all__ = ["bary_batch", "bary_recursive"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pybary']

package_data = \
{'': ['*']}

install_requires = \
['ipykernel>=6.20.1,<7.0.0',
 'ipython>=8.8.0,<9.0.0',
 'jupyter-core>=5.1.2,<6.0.0',
 'matplotlib>=3.6.2,<4.0.0',
 'numpy==1.24.1',
 'seaborn>=0.12.2,<0.13.0']

setup_kwargs = {
    'name': 'pybary',
    'version': '0.1.8',
    'description': 'Barycenter method in python',
    'long_description': '[![Version](https://img.shields.io/pypi/v/pybary.svg)](https://pypi.python.org/pypi/pybary)\n[![python](https://img.shields.io/pypi/pyversions/pybary.svg)](https://pypi.org/project/pybary/)\n[![codecov](https://codecov.io/gh/asmove/pybary/branch/main/graph/badge.svg?token=G8TVJ4X32L)](https://codecov.io/gh/asmove/pybary)\n[![downloads](https://img.shields.io/pypi/dm/pybary)](https://pypi.org/project/pybary/)\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/asmove/pybary/HEAD)\n\nPybary\n========\n\n![A sniffer optimizer](https://github.com/asmove/pybary/blob/main/images/pybary-tiny.png?raw=true)\n\nBarycenter method in python. Take a look at original article: https://arxiv.org/abs/1801.10533\n\nHow to install\n----------------\n\nWe run the command on desired installation environment:\n\n``` {.bash}\npip install pybary\n```\n\nMinimal example\n----------------\n\nWe may code examples by performing following actions \n\n1. Run command `python examples/example.py` from package root folder;\n2. Open notebook `examples/example.ipynb` and run cell on given order.\n',
    'author': 'Bruno Peixoto',
    'author_email': 'brunolnetto@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/pybary/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<3.12',
}


setup(**setup_kwargs)
